@@ -1,7 +1,7 @@
 /*
  * Univention Directory Notifier
  *
- * Copyright 2004-2020 Univention GmbH
+ * Copyright 2004-2021 Univention GmbH
  *
  * https://www.univention.de/
  *
@@ -42,6 +42,7 @@
 #include <time.h>
 #include <ldap.h>
 #include <sasl/sasl.h>
+#include <assert.h>
 #include <univention/debug.h>
 
 #include "notify.h"
@@ -56,6 +57,7 @@ extern NotifyId_t notify_last_id;
 extern Notify_t notify;
 extern long long notifier_lock_count;
 extern long long notifier_lock_time;
+extern void set_listener_callback(int sig, siginfo_t *si, void *data);
 extern void unset_listener_callback ();
 
 extern unsigned long SCHEMA_ID;
@@ -67,7 +69,7 @@ static FILE* fopen_lock(const char *name, const char *type)
 	int fd;
 
 	if ((file = fopen(name, type)) == NULL) {
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "ERROR Could not open file [%s]\n", name);
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "ERROR Could not open file [%s]", name);
 		return NULL;
 	}
 
@@ -76,10 +78,10 @@ static FILE* fopen_lock(const char *name, const char *type)
 		int rc = lockf(fd, F_TLOCK, 0);
 		if (!rc)
 			break;
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "Could not get lock for file [%s]; count=%d\n", name, count);
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "Could not get lock for file [%s]; count=%d", name, count);
 		count++;
 		if (count > notifier_lock_count) {
-			univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ERROR, "Could not get lock for file [%s]; exit\n", name);
+			univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ERROR, "Could not get lock for file [%s]; exit", name);
 			exit(0);
 		}
 		usleep(notifier_lock_time);
@@ -96,7 +98,7 @@ static FILE* fopen_dotlockfile(const char *name)
 	snprintf( buf, sizeof(buf), "%s.lock", name );
 
 	if ((file = fopen_lock(buf, "a")) == NULL) {
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "ERROR Could not open lock file [%s]\n", buf);
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "ERROR Could not open lock file [%s]", buf);
 		return NULL;
 	}
 
@@ -119,7 +121,7 @@ static FILE* fopen_with_lockfile(const char *name, const char *type, FILE **l_fi
 	}
 
 	if ((file = fopen(name, type)) == NULL) {
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "ERROR Could not open file [%s]\n", name);
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "ERROR Could not open file [%s]", name);
 
 		int l_fd = fileno(*l_file);
 		lockf(l_fd, F_ULOCK, 0);
@@ -164,116 +166,52 @@ static int fclose_with_lockfile(const char *name, FILE **file, FILE **l_file)
 	return 0;
 }
 
-/* Allocate and initialize a new entry. */
-static NotifyEntry_t *notify_entry_alloc()
+
+/* Parse "<id> <dn with blanks> <command>\n" */
+static int parse_transaction_line(NotifyEntry_t* entry, char* line)
 {
-	NotifyEntry_t *entry = malloc(sizeof(NotifyEntry_t));
-	if (entry)
-		notify_entry_init(entry);
-	return entry;
+	char *first_space_p = index(line, ' ');
+	char *last_space_p = rindex(line, ' ');
+	assert(first_space_p && last_space_p && first_space_p < last_space_p);
+	size_t size = last_space_p - first_space_p - 1;
+	entry->command = last_space_p[1];
+	entry->dn = strndup(first_space_p + 1, size);
+	return sscanf(line, "%ld", &(entry->notify_id.id)) == 1 && entry->notify_id.id > 0 && entry->dn;
 }
 
-static long split_transaction_buffer ( NotifyEntry_t *entry, char *buf, long l_buf)
+static void notify_dump_to_files( Notify_t *notify, NotifyEntry_t *entry)
 {
-	NotifyEntry_t *tmp=NULL;
-	NotifyEntry_t *tmp2=NULL;
-
-	int size;
-
-	int start=1;
-
-	char *p, *p1;
-
-	char *s;
-
-	char *p_tmp1, *p_tmp2;
-
-	tmp2=entry;
-
-	p1=strndup(buf, l_buf);
-	p=p1;
-
-	for ( s=strtok(p,"\n"); s!= NULL; s=strtok(NULL,"\n") ) {
-		if ( start ) {
-			sscanf(s, "%ld", &(tmp2->notify_id.id));
-			tmp2->command=s[strlen(s)-1];
-			p_tmp1=index(s, ' ');
-			p_tmp2=rindex(s, ' ');
-			size=p_tmp2-p_tmp1;
-			tmp2->dn=malloc((size)*sizeof(char));
-			memcpy( tmp2->dn, p_tmp1+1, p_tmp2-p_tmp1);
-			tmp2->dn[size-1]='\0';
-
-			tmp2->next=NULL;
-		} else {
-			tmp = notify_entry_alloc();
-
-			sscanf(s, "%ld", &(tmp->notify_id.id));
-			tmp->command=s[strlen(s)-1];
-			p_tmp1=index(s, ' ');
-			p_tmp2=rindex(s, ' ');
-			size=p_tmp2-p_tmp1;
-			tmp->dn=malloc((size)*sizeof(char));
-			memcpy( tmp->dn, p_tmp1+1, p_tmp2-p_tmp1);
-			tmp->dn[size-1]='\0';
-
-			tmp->next=NULL;
-			tmp2->next=tmp;
-			tmp2=tmp;
-		}
-
-		start=0;
-	}
-
-	free(p1);
-	return 0;
-}
-
-void notify_dump_to_files( Notify_t *notify, NotifyEntry_t *entry)
-{
-	NotifyEntry_t *tmp;
 	char buffer[2048];
 	FILE *index = NULL;
 
-	if (entry == NULL)
-		return;
-
-
 	univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ALL, "LOCK from dump_to_files");
 	if ((notify->tf = fopen_with_lockfile(FILE_NAME_TF, "a", &(notify->l_tf))) == NULL) {
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "ERROR on open tf\n");
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "ERROR on open tf");
 		goto error;
 	}
 	if ((index = index_open(FILE_NAME_TF_IDX)) == NULL) {
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "unable to open index\n");
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "unable to open index");
 		goto error;
 	}
 
-	for (tmp = entry; tmp != NULL; tmp = tmp->next) {
-		if (tmp->dn != NULL && tmp->notify_id.id >= 0) {
-			long offset = ftell(notify->tf);
-			int len = snprintf(buffer, sizeof(buffer), "%ld %s %c\n", tmp->notify_id.id, tmp->dn, tmp->command);
-			if (len >= sizeof(buffer)) {
-				univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ERROR, "buffer too small");
-				abort();
-			}
-
-			index_set(index, tmp->notify_id.id, offset);
-			univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "want to write to transaction file; id=%ld", tmp->notify_id.id);
-			if (fallocate(fileno(notify->tf), FALLOC_FL_KEEP_SIZE, offset, len) == -1 && (errno != ENOSYS) && (errno != EOPNOTSUPP)) {
-				perror("Failed fallocate(tf)");
-				abort();
-			}
-			if (fprintf(notify->tf, "%s", buffer) != len) {
-				perror("Failed fprintf(tf)");
-				abort();
-			}
-			univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "wrote to transaction file; id=%ld; dn=%s, cmd=%c", tmp->notify_id.id, tmp->dn, tmp->command);
-		} else {
-			univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "tmp->dn == NULL; id=%ld", tmp->notify_id.id);
-		}
+	long offset = ftell(notify->tf);
+	int len = snprintf(buffer, sizeof(buffer), "%ld %s %c\n", entry->notify_id.id, entry->dn, entry->command);
+	if (len >= sizeof(buffer)) {
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ERROR, "buffer too small");
+		abort();
 	}
-	univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "wrote to transaction file; close");
+
+	index_set(index, entry->notify_id.id, offset);
+	univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "want to write to transaction file; id=%ld", entry->notify_id.id);
+	if (fallocate(fileno(notify->tf), FALLOC_FL_KEEP_SIZE, offset, len) == -1 && (errno != ENOSYS) && (errno != EOPNOTSUPP)) {
+		perror("Failed fallocate(tf)");
+		abort();
+	}
+	if (fprintf(notify->tf, "%s", buffer) != len) {
+		perror("Failed fprintf(tf)");
+		abort();
+	}
+	univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "wrote to transaction file; id=%ld; dn=%s, cmd=%c", entry->notify_id.id, entry->dn, entry->command);
 
 error:
 	if (index)
@@ -320,9 +258,6 @@ static void notify_dump_to_ldap(NotifyEntry_t *trans) {
 		.sa_flags = 0,
 	};
 
-	if (trans == NULL)
-		return;
-
 	// libldap uses liblber uses write() on the SOCKET to slapd, which raises SIGPIPE when slapd closes the socket due to a timeout.
 	sigaction(SIGPIPE, &act, &oldact);
 
@@ -349,89 +284,82 @@ reopen:
 		}
 	}
 
-	for (; trans != NULL; trans = trans->next) {
-		if (!trans->dn)
-			continue;
-		if (!trans->notify_id.id)
-			continue;
+	char dn[44]; // strlen("reqSession=%ld,cn=translog") + strlen(ULONG_MAX)
+	snprintf(dn, sizeof(dn), "reqSession=%ld,cn=translog", trans->notify_id.id);
 
-		char dn[44];  // strlen("reqSession=%ld,cn=translog") + strlen(ULONG_MAX)
-		snprintf(dn, sizeof(dn), "reqSession=%ld,cn=translog", trans->notify_id.id);
+	char *oc_values[] = { "auditObject", NULL };
+	LDAPMod oc_mod = {
+		.mod_op = LDAP_MOD_ADD,
+		.mod_type = "objectClass",
+		.mod_values = oc_values,
+	};
 
-		char *oc_values[] = { "auditObject", NULL };
-		LDAPMod oc_mod = {
-			.mod_op = LDAP_MOD_ADD,
-			.mod_type = "objectClass",
-			.mod_values = oc_values,
-		};
+	char start[16]; // strlen('YYYYmmddHHMMSSZ')
+	time_t t = time(NULL);
+	struct tm tm ;
+	gmtime_r(&t, &tm);
+	strftime(start, sizeof(start), "%Y%m%d%H%M%SZ", &tm);
+	char *start_values[] = { start, NULL };
+	LDAPMod time_mod = {
+		.mod_op = LDAP_MOD_ADD,
+		.mod_type = "reqStart",
+		.mod_values = start_values,
+	};
 
-		char start[16]; // strlen('YYYYmmddHHMMSSZ')
-		time_t t = time(NULL);
-		struct tm tm ;
-		gmtime_r(&t, &tm);
-		strftime(start, sizeof(start), "%Y%m%d%H%M%SZ", &tm);
-		char *start_values[] = { start, NULL };
-		LDAPMod time_mod = {
-			.mod_op = LDAP_MOD_ADD,
-			.mod_type = "reqStart",
-			.mod_values = start_values,
-		};
+	char id[21];  // strlen(ULONG_MAX)
+	snprintf(id, sizeof(id), "%ld", trans->notify_id.id);
+	char *index_values[] = { id, NULL };
+	LDAPMod index_mod = {
+		.mod_op = LDAP_MOD_ADD,
+		.mod_type = "reqSession",
+		.mod_values = index_values,
+	};
 
-		char id[21];  // strlen(ULONG_MAX)
-		snprintf(id, sizeof(id), "%ld", trans->notify_id.id);
-		char *index_values[] = { id, NULL };
-		LDAPMod index_mod = {
-			.mod_op = LDAP_MOD_ADD,
-			.mod_type = "reqSession",
-			.mod_values = index_values,
-		};
+	char *dn_values[] = { trans->dn, NULL };
+	LDAPMod dn_mod = {
+		.mod_op = LDAP_MOD_ADD,
+		.mod_type = "reqDN",
+		.mod_values = dn_values,
+	};
 
-		char *dn_values[] = { trans->dn, NULL };
-		LDAPMod dn_mod = {
-			.mod_op = LDAP_MOD_ADD,
-			.mod_type = "reqDN",
-			.mod_values = dn_values,
-		};
+	char cmd[2];
+	cmd[0] = trans->command;
+	cmd[1] = '\0';
+	char *cmd_values[] = { cmd, NULL };
+	LDAPMod cmd_mod = {
+		.mod_op = LDAP_MOD_ADD,
+		.mod_type = "reqType",
+		.mod_values = cmd_values,
+	};
 
-		char cmd[2];
-		cmd[0] = trans->command;
-		cmd[1] = '\0';
-		char *cmd_values[] = { cmd, NULL };
-		LDAPMod cmd_mod = {
-			.mod_op = LDAP_MOD_ADD,
-			.mod_type = "reqType",
-			.mod_values = cmd_values,
-		};
+	LDAPMod *attrs[] = {
+		&oc_mod,
+		&time_mod,
+		&index_mod,
+		&dn_mod,
+		&cmd_mod,
+		NULL
+	};
 
-		LDAPMod *attrs[] = {
-			&oc_mod,
-			&time_mod,
-			&index_mod,
-			&dn_mod,
-			&cmd_mod,
-			NULL
-		};
+	univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ALL, "LDIF dn: %s", dn);
+	for (rc = 0; attrs[rc]; rc++)
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ALL, "LDIF %s: %s", attrs[rc]->mod_type, attrs[rc]->mod_values[0]);
 
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ALL, "LDIF dn: %s", dn);
-		for (rc = 0; attrs[rc]; rc++)
-			univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ALL, "LDIF %s: %s", attrs[rc]->mod_type, attrs[rc]->mod_values[0]);
-
-		rc = ldap_add_ext_s(ld, dn, attrs, serverctrls, clientctrls);
-		switch (rc) {
-			case LDAP_SUCCESS:
-				break;
-			case LDAP_SERVER_DOWN:
-				if ((rc = ldap_unbind_ext_s(ld, serverctrls, clientctrls)) != LDAP_SUCCESS)
-					univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_PROCESS, "%ld ldap_unbind_ext_s(): %s", trans->notify_id.id, ldap_err2string(rc));
-				ld = NULL;
-				goto reopen;
-			case LDAP_ALREADY_EXISTS:
-				univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "%ld ldap_add() already exists", trans->notify_id.id);
-				break;
-			default:
-				univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ERROR, "%ld ldap_add(): %s", trans->notify_id.id, ldap_err2string(rc));
-				abort();
-		}
+	rc = ldap_add_ext_s(ld, dn, attrs, serverctrls, clientctrls);
+	switch (rc) {
+		case LDAP_SUCCESS:
+			break;
+		case LDAP_SERVER_DOWN:
+			if ((rc = ldap_unbind_ext_s(ld, serverctrls, clientctrls)) != LDAP_SUCCESS)
+				univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_PROCESS, "%ld ldap_unbind_ext_s(): %s", trans->notify_id.id, ldap_err2string(rc));
+			ld = NULL;
+			goto reopen;
+		case LDAP_ALREADY_EXISTS:
+			univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "%ld ldap_add() already exists", trans->notify_id.id);
+			break;
+		default:
+			univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ERROR, "%ld ldap_add(): %s", trans->notify_id.id, ldap_err2string(rc));
+			abort();
 	}
 
 	sigaction(SIGPIPE, &oldact, NULL);
@@ -443,10 +371,6 @@ void notify_init ( Notify_t *notify )
 	notify->l_tf  = NULL;
 }
 
-void notify_entry_init(NotifyEntry_t *entry)
-{
-	memset(entry, 0, sizeof(NotifyEntry_t));
-}
 
 int notify_transaction_get_last_notify_id ( Notify_t *notify, NotifyId_t *notify_id )
 {
@@ -455,7 +379,7 @@ int notify_transaction_get_last_notify_id ( Notify_t *notify, NotifyId_t *notify
 
 	univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ALL, "LOCK from notify_transaction_get_last_notify_id");
 	if ((notify->tf = fopen_with_lockfile(FILE_NAME_TF, "r", &(notify->l_tf))) == NULL) {
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "unable to lock notify_id\n");
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "unable to lock notify_id");
 		notify_id->id = 0;
 		return -1;
 	}
@@ -469,13 +393,10 @@ int notify_transaction_get_last_notify_id ( Notify_t *notify, NotifyId_t *notify
 	if ( c == -1 || c == 255 ) {
 		/* empty file */
 		notify_id->id = 0;
-
 	} else if ( ftell(notify->tf) == 1) {
-
 		/* only one entry */
 		fseek( notify->tf, 0, SEEK_SET);
 		fscanf(notify->tf, "%ld",& (notify_id->id));
-
 	} else {
 		fscanf(notify->tf, "%ld",& (notify_id->id));
 	}
@@ -485,37 +406,24 @@ int notify_transaction_get_last_notify_id ( Notify_t *notify, NotifyId_t *notify
 	return 0;
 }
 
-void notify_entry_free(NotifyEntry_t *entry )
-{
-	NotifyEntry_t *tmp = entry;
-	NotifyEntry_t *tmp2;
-
-	while ( tmp != NULL ) {
-		if ( tmp->dn) free(tmp->dn);
-		tmp2=tmp;
-		tmp=tmp->next;
-		free(tmp2);
-	}
-}
 
 char* notify_transcation_get_one_dn ( unsigned long last_known_id )
 {
 	char buffer[2048];
-	int i, size;
+	int i;
 	char c;
 	unsigned long id;
 	bool found = false;
 	FILE *index = NULL;
 	size_t pos;
 
-
 	univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ALL, "LOCK from notify_transcation_get_one_dn");
 	if ((notify.tf = fopen_with_lockfile(FILE_NAME_TF, "r", &(notify.l_tf))) == NULL) {
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "unable to lock tf\n");
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "unable to lock tf");
 		return NULL;
 	}
 	if ( ( index = index_open(FILE_NAME_TF_IDX) ) == NULL ) {
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "unable to open index\n");
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_WARN, "unable to open index");
 		goto error;
 	}
 
@@ -532,7 +440,7 @@ char* notify_transcation_get_one_dn ( unsigned long last_known_id )
 			sscanf(buffer, "%ld", &id);
 			if (id == last_known_id) {
 				found = true;
-				univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "Found (get_one_dn, index) %ld",id);
+				univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "Found (get_one_dn, index) %ld", id);
 			}
 		}
 	}
@@ -545,13 +453,13 @@ char* notify_transcation_get_one_dn ( unsigned long last_known_id )
 		}
 
 		if ( c == '\n' ) {
-			size = sscanf(buffer, "%ld", &id) ;
+			sscanf(buffer, "%ld", &id) ;
 
 			index_set(index, id, pos);
 
 			if ( id == last_known_id ) {
 				found = true;
-				univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "Found (get_one_dn) %ld",id);
+				univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "Found (get_one_dn) %ld", id);
 				break;
 			}
 
@@ -577,36 +485,6 @@ error:
 	}
 }
 
-char* notify_entry_to_string(NotifyEntry_t entry )
-{
-	int len = 0;
-	char *str, *p;
-	char buffer[32];
-	int rc;
-
-	if ( entry.dn == NULL ) {
-		return NULL;
-	}
-
-	len += 4; /* space + space + newline */
-	len += strlen(entry.dn);
-	len += snprintf(buffer,32, "%ld",entry.notify_id.id);
-
-	len+=1;
-	if ( (str = malloc(len*sizeof(char) ) ) == NULL ) {
-		return NULL;
-	}
-
-	memset(str, 0, len);
-	p=str;
-
-	rc = sprintf(p, "%ld %s %c\n", entry.notify_id.id, entry.dn, entry.command);
-	p+=rc;
-
-	return str;
-
-}
-
 void notify_schema_change_callback(int sig, siginfo_t *si, void *data)
 {
 	FILE *file;
@@ -621,29 +499,19 @@ void notify_schema_change_callback(int sig, siginfo_t *si, void *data)
 	univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_INFO, "NEW Schema ID = %ld", SCHEMA_ID);
 
 	fclose(file);
-
 }
 
 void notify_listener_change_callback(int sig, siginfo_t *si, void *data)
 {
-	NotifyEntry_t *entry = NULL;
-
+	NotifyEntry_t entry = {0,};
 	FILE *file, *l_file;
 	int fd;
 	int rc;
-
 	struct stat stat_buf;
 
-	char *buf = NULL;
-
-	int nread;
-
-
 	univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ALL, "NOTIFY Listener" );
 	univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ALL, "... go");
 
-	entry = notify_entry_alloc();
-
 	univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ALL, "LOCK from notify_listener_change_callback");
 	rc = stat(FILE_NAME_NOTIFIER_PRIV, &stat_buf);
 	if ( (rc == -1) || (stat_buf.st_size == 0) ) {
@@ -665,64 +533,39 @@ void notify_listener_change_callback(int sig, siginfo_t *si, void *data)
 			close(fd);
 		}
 
-               unset_listener_callback();
+		unset_listener_callback();
 		fclose_lock(&l_file);
 	} else {
-		set_listener_callback();
+		set_listener_callback(0, NULL, NULL);
 	}
 
 	if ( ( file = fopen_lock (FILE_NAME_NOTIFIER_PRIV, "r+") ) == NULL ) {
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ERROR, "Could not open %s\n", FILE_NAME_NOTIFIER_PRIV);
+		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ERROR, "Could not open %s", FILE_NAME_NOTIFIER_PRIV);
 	}
 
-	fd = fileno(file);
-	if( ( fstat (fd, &stat_buf) ) != 0 ) {
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ERROR, "stat error\n");
-		goto error;
-	}
-
-	if (stat_buf.st_size == 0) {
-		goto error;
-	}
-
-	if ( (buf = malloc( (stat_buf.st_size + 1 ) * sizeof(char))) == NULL ) {
-		univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ERROR, "malloc error\n");
-		goto error;
-	}
-
-	memset(buf, 0, stat_buf.st_size + 1);
-
-	if ( (nread = fread( buf, sizeof(char), stat_buf.st_size, file)) != 0 ) {
-		split_transaction_buffer ( entry, buf, nread);
-		notify_dump_to_ldap(entry);
-		notify_dump_to_files(&notify, entry);
-		unlink(FILE_NAME_NOTIFIER_PRIV);
-
-		{
-			NotifyEntry_t *tmp;
-			char *dn_string = NULL;
-
-			tmp = entry;
-
-			while ( tmp != NULL ) {
-				notifier_cache_add(tmp->notify_id.id, tmp->dn, tmp->command);
-				notify_last_id.id=tmp->notify_id.id;
-				dn_string = notify_entry_to_string( *tmp );
-				if ( dn_string != NULL ) {
-					network_client_all_write(tmp->notify_id.id, dn_string, strlen(dn_string) );
-				}
-				tmp=tmp->next;
-			}
-			free(dn_string);
+	char *line = NULL;
+	size_t len = 0;
+	while (getline(&line, &len, file) != -1) {
+		if (!parse_transaction_line(&entry, line)) {
+			univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_ERROR, "ABORTING EXECUTION: Invalid transaction line %s of file %s", line, FILE_NAME_NOTIFIER_PRIV);
+			free(entry.dn);
+			free(line);
+			fclose_lock(&file);
+			abort();
 		}
+		if (entry.notify_id.id <= notify_last_id.id) {
+			univention_debug(UV_DEBUG_TRANSFILE, UV_DEBUG_PROCESS, "Skipping old Transaction ID %ld before %ld from %s", entry.notify_id.id, notify_last_id.id, FILE_NAME_NOTIFIER_PRIV);
+			continue;
+		}
+		notify_dump_to_ldap(&entry);
+		notify_dump_to_files(&notify, &entry);
+		notifier_cache_add(entry.notify_id.id, entry.dn, entry.command);
+		notify_last_id.id = entry.notify_id.id;
+		network_client_all_write(entry.notify_id.id, line, strlen(line));
+		free(entry.dn);
 	}
 
-	notify_entry_free(entry);
-
-	free(buf);
-
-error:
+	unlink(FILE_NAME_NOTIFIER_PRIV);
+	free(line);
 	fclose_lock(&file);
-
-	return;
 }
